from inspect import iscoroutinefunction
from abc import ABC, abstractmethod
from typing import Callable, Iterable, Self

from .error_info import check_bad_value
from .callback_data import CallbackData
from .message import HasButtonRows, Message, SentMessage

class HasCallbackData(ABC):
    @abstractmethod
    def __init__(self):
        self.messages: list[HasButtonRows]
    
    def get_callback_data(self):
        result: list[CallbackData] = []
        for message in self.messages:
            result.extend(message.get_callback_data())
        return result

class ReadyScreen(HasCallbackData):
    def __init__(self, *messages: Message):
        self.messages: list[Message] = []
        self.extend(list(messages))
    
    def extend(self, messages: list[Message]):
        for message in messages:
            self.append(message)
        
    def append(self, message: Message):
        check_bad_value(message, Message, self, "message")
        self.messages.append(message)
    
    def __repr__(self):
        return f"{type(self).__name__}({self.messages!r})"
    
    def clone(self) -> "ReadyScreen":
        return ReadyScreen(*[message.clone() for message in self.messages])

class SentScreen(HasCallbackData):
    def __init__(self, *messages: SentMessage):
        self.messages: list[SentMessage] = []
        self.extend(list(messages))
    
    def extend(self, messages: list[SentMessage]):
        for message in messages:
            self.append(message)
        
    def append(self, message: SentMessage):
        check_bad_value(message, SentMessage, self, "message")
        self.messages.append(message)
        
    def clone(self) -> Self:
        return self.__class__(*[message.clone() 
                                for message in self.messages])
    
    def __repr__(self):
        return f"{type(self).__name__}({self.messages!r})"
    
    @abstractmethod
    async def delete(self, *args, **kwargs): ...
    
    @abstractmethod
    def get_unsent(self) -> ReadyScreen: ...

class ProtoScreen(ABC):
    def __init__(self, name: str = None):
        self.name = name
        self.messages: list[Message] = []
    
    def append(self, message: Message):
        check_bad_value(message, Message, self, "message")
        self.messages.append(message)
    
    def extend(self, messages: list[Message]):
        for message in messages:
            self.append(message)
    
    @abstractmethod
    async def evaluate(self, user_id: int, *args, 
                       **kwargs) -> ReadyScreen: ...

class StaticScreen(ProtoScreen):
    def __init__(self, name: str, *messages: Message):
        super().__init__(name = name)
        self.extend(list(messages))
    
    async def evaluate(self, user_id: int, *args, **kwargs):
        messages = []
        for message in self.messages:
            new_message = message.clone()
            messages.append(new_message)
        return ReadyScreen(*messages)
    
    def __repr__(self):
        return f"{type(self).__name__}({self.name!r}, {self.messages!r})"

class DynamicScreen(ProtoScreen):
    def __init__(self, name: str, 
            function: Callable[[int], Iterable[Message]]):
        super().__init__(name)
        if not iscoroutinefunction(function):
            print(f"Экран {name} был создан с не async функцией")
        self.function = function
    
    async def evaluate(self, user_id: int, **kwargs):
        messages = await self.function(user_id=user_id, **kwargs)
        return ReadyScreen(*messages)

    def __repr__(self):
        return f"{type(self).__name__}({self.name!r}, {self.function!r})"




from asyncio import gather
from telegram import Bot

from ..screen import ReadyScreen
from .screen import SentScreen
from ..user_data import UserData
from ..user_screen import UserScreen as BaseUserScreen
from .messages.message import SentMessage, Message

class UserScreen(BaseUserScreen):
    def __init__(self, user_data: UserData, bot: Bot):
        super().__init__(user_data)
        self.bot = bot
    
    async def clear(self, user_id: int, delete_messages: bool = True):
        user_data = self.user_data.get(user_id)
        screen = user_data.screen
        if screen and delete_messages:
            await screen.delete(self.bot)
        user_data.screen = None
    
    async def set(self, user_id: int, new_screen: ReadyScreen = None):
        if new_screen == None:
            new_screen = ReadyScreen()
        mapping = self._map_callback_data(user_id, new_screen)
        
        old_screen = self.get(user_id)
        user_data = self.user_data.get(user_id)
        delete, edit, send = self.calc_screen_difference(
            old_screen, new_screen, 
            Message, SentMessage)
        
        new_screen = SentScreen()
        tasks = []
        for message in delete:
            tasks.append( message.delete(self.bot) )
            
        for old_message, new_message in edit:
            transformed_message = new_message.transform(old_message)
            tasks.append( transformed_message.edit(self.bot, mapping) )
            new_screen.extend([transformed_message])
        
        for message in send:
            new_message = await message.send(user_id, self.bot, mapping)
            new_screen.extend([new_message])
            
        await gather(*tasks)
        
        user_data.screen = new_screen
    
    async def buffer(self, user_id: int):
        user_data = self.user_data.get(user_id)
        unsent = None
        if user_data.screen:
            unsent = user_data.screen.get_unsent()
            await user_data.screen.delete(self.bot)
        user_data.screen_buffer = unsent
        user_data.screen = None
        


from telegram import Message as PTBMessage
from ...message import VideoNoteMessage as BaseVideoNoteMessage
from ...message import SentVideoNoteMessage as BaseSentVideoNoteMessage
from .message import HasButtonRows, Message, SentMessage

class VideoNoteMessage(BaseVideoNoteMessage, Message): ...

class SentVideoNoteMessage(BaseSentVideoNoteMessage, SentMessage): ...


from types import FunctionType, MethodType
from typing import Callable, Self, Sequence, Type
from abc import ABC, abstractmethod
from .session import InputSession
from .input_callback import FuncCallback, check_pre_post_func
from .error_info import check_bad_text_and_len, check_bad_value, check_callable

class CallbackData(ABC):
    @abstractmethod
    def clone(self) -> Self: ...
    
    @abstractmethod
    def __repr__(self) -> str: ...
    
    @abstractmethod
    async def use(self, *, user_id: int,
                  input_sessions: Sequence[InputSession],
                  screen_set_by_name: Callable, 
                  screen_step_back: Callable, 
                  reset_input_callback: Callable,
                  update_sessions: Callable,
                  **kwargs):
        ...

class Dummy(CallbackData):
    def clone(self):
        return Dummy()
    
    def __repr__(self):
        return f"{type(self).__name__}()"
    
    async def use(self, *, user_id: int,
                  input_sessions: Sequence[InputSession],
                  screen_set_by_name: Callable, 
                  screen_step_back: Callable, 
                  reset_input_callback: Callable,
                  update_sessions: Callable,
                  **kwargs):
        pass

class RunFunc(CallbackData):
    def __init__(self, function: Callable, **kwargs):
        """Использование:  
            function - Функция для выполнения при нажатии кнопки  
            **kwargs - keyword аргументы функции
        """
        check_callable(function, self, "function")
        self.function = function
        self.kwargs = kwargs
    
    def clone(self):
        return RunFunc(self.function, **self.kwargs)
    
    def __repr__(self):
        return f"{type(self).__name__}({self.function!r})"
    
    def __eq__(self, other: object):
        return isinstance(other, RunFunc) and \
            self.function == other.function and self.kwargs == other.kwargs
            
    async def use(self, *, user_id: int,
                  input_sessions: Sequence[InputSession],
                  screen_set_by_name: Callable, 
                  screen_step_back: Callable, 
                  reset_input_callback: Callable,
                  update_sessions: Callable,
                  **kwargs):
        await self.function(user_id=user_id, **self.kwargs, **kwargs)

class GoToScreen(CallbackData):
    def __init__(self, screen_name: str, *
            , pre_func:  FuncCallback | None = None
            , post_func: FuncCallback | None = None):
        check_bad_text_and_len(screen_name, self, "screen_name")
        check_pre_post_func(pre_func, post_func, self)
        
        self.screen_name = screen_name
        self.pre_func = pre_func
        self.post_func = post_func
    
    def clone(self):
        return GoToScreen(self.screen_name)
    
    def __repr__(self):
        return f"{type(self).__name__}({self.screen_name!r}, \
{self.pre_func}, {self.post_func})"
    
    def __eq__(self, other: object):
        return isinstance(other, GoToScreen) and \
            self.screen_name == other.screen_name

    async def use(self, *, user_id: int,
                  input_sessions: Sequence[InputSession],
                  screen_set_by_name: Callable, 
                  screen_step_back: Callable, 
                  reset_input_callback: Callable,
                  update_sessions: Callable,
                  **kwargs):
        if self.pre_func:
            await self.pre_func(user_id=user_id, **kwargs)
        
        await screen_set_by_name(user_id, self.screen_name, 
                                 **kwargs)
        
        if self.post_func:
            await self.post_func(user_id=user_id, **kwargs)
            
        update_sessions()

class StepBack(CallbackData):
    def __init__(self, times: int = 1, clear_input_callback: bool = True
            , pop_last_input: bool = True
            , pre_func: FuncCallback | None = None
            , post_func: FuncCallback | None = None):

        check_bad_value(times, int, self, "times")
        self.times = times

        check_bad_value(clear_input_callback, bool, self, "clear_input_callback")
        self.clear_input_callback = clear_input_callback

        check_bad_value(pop_last_input, bool, self, "pop_last_input")
        self.pop_last_input = pop_last_input

        check_pre_post_func(pre_func, post_func, self)
        
        self.pre_func = pre_func
        self.post_func = post_func

    def clone(self):
        return StepBack()
    
    def __repr__(self):
        return f"{type(self).__name__}({self.times!r}, {self.clear_input_callback}, \
{self.pop_last_input}, {self.pre_func}, {self.post_func})"
    
    def __eq__(self, other: object):
        return isinstance(other, StepBack)
    
    async def use(self, *, user_id: int,
                  input_sessions: Sequence[InputSession],
                  screen_set_by_name: Callable, 
                  screen_step_back: Callable, 
                  reset_input_callback: Callable,
                  update_sessions: Callable,
                  **kwargs):
        if self.clear_input_callback:
            reset_input_callback()
        
        for session in input_sessions:
            if self.pop_last_input and session.may_pop_last_input:
                for i in range(self.times):
                    if session.messages == []:
                        break
                    session.messages.pop()
        
        if self.pre_func:
            await self.pre_func(user_id=user_id, **kwargs)
        
        await screen_step_back(user_id, self.times)
        
        if self.post_func:
            await self.post_func(user_id=user_id, **kwargs)
        
        update_sessions()

class CallbackDataMapping:
    def __init__(self):
        self.items: list[tuple[CallbackData, str]] = []
    
    def add(self, callback: CallbackData, uuid: str):
        self.items.append((callback, uuid))
    
    def get_by_callback(self, callback: CallbackData):
        for i_callback, uuid in self.items:
            if callback is i_callback:
                return uuid
        raise KeyError(callback)
    
    def get_by_uuid(self, uuid: str):
        for callback, i_uuid in self.items:
            if uuid == i_uuid:
                return callback
        return None


from abc import abstractmethod
from typing import Self
from .callback_data import CallbackData
from .error_info import check_bad_text_and_len, check_bad_value

class Button:
    def __init__(self, text: str, callback_data: CallbackData, 
                 url: str | None = None, web_app = None):
        check_bad_text_and_len(text, self, "text")
        check_bad_value(callback_data, CallbackData, self, "callback_data")
        if url:
            check_bad_text_and_len(url, self, "url")

        self.text = text
        self.callback_data = callback_data
        self.url = url
        self.web_app = web_app
    
    def clone(self) -> Self:
        return self.__class__(self.text, self.callback_data.clone(), 
                      self.url, self.web_app)

    def __repr__(self):
        return f"{type(self).__name__}({self.text!r}, {self.callback_data!r}, url={self.url!r})"

    def __eq__(self, other: object):
        if not isinstance(other, Button):
            return False
        
        return (self.text == other.text and \
                self.callback_data == other.callback_data and \
                self.url == other.url and \
                self.web_app == other.web_app)

class ButtonRow:
    def __init__(self, *buttons: Button):
        self.buttons: list[Button] = []
        self.extend(list(buttons))
    
    def extend(self, buttons: list[Button]):
        for button in buttons:
            self.append(button)
        return self

    def append(self, button: Button):
        check_bad_value(button, Button, self, "button")
        self.buttons.append(button)
        return self
    
    def clone(self) -> Self:
        return self.__class__().extend([
            button.clone() 
            for button in self.buttons
        ])
    
    def __repr__(self):
        return f"{type(self).__name__}(*{self.buttons!r})"
    
    def __eq__(self, other: object) -> bool:
        return \
            isinstance(other, self.__class__) and \
            all([
                button1 == button2 
                for button1, button2
                in zip(self.buttons, other.buttons)
            ])

class ButtonRows:
    def __init__(self, *rows: ButtonRow):
        self.rows: list[ButtonRow] = []
        self.extend(list(rows))
    
    def extend(self, rows: list[ButtonRow]):
        for row in rows:
            self.append(row)
    
    def append(self, row: ButtonRow):
        check_bad_value(row, ButtonRow, self, "row")
        self.rows.append(row)
    
    def clone(self) -> Self:
        return self.__class__(*[row.clone() for row in self.rows])
    
    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and \
            all([row1 == row2
                for row1, row2 in 
                zip(self.rows,other.rows)])
        
    def __repr__(self):
        return f"{type(self).__name__}(*{self.rows!r})"
    
    @abstractmethod
    def to_reply_markup(self): ...
    
    def get_callback_data(self):
        result: list[CallbackData] = []
        for row in self.rows:
            for button in row.buttons:
                result.append(button.callback_data)
        return result


from typing import Any

class Session:
    def __init__(self, id: str, delete_if_level_decreased = True, 
            delete_if_last_dir_changed = False):

        self.id = id
        
        self.directory_level: int = -1
        self.last_directory: str = ""
        
        self.delete_if_level_decreased = delete_if_level_decreased
        self.delete_if_last_dir_changed = delete_if_last_dir_changed
    
    def __repr__(self):
        return f"{type(self).__name__}(id={self.id!r})"

class InputSession(Session):
    def __init__(self, id: str, delete_if_level_decreased = True, 
            delete_if_last_dir_changed = False,add_new_messages = True, 
            may_pop_last_input = True):
        
        super().__init__(id, delete_if_level_decreased, delete_if_last_dir_changed)
        
        self.messages = []
        self.add_new_messages = add_new_messages
        self.may_pop_last_input = may_pop_last_input
    
    def append(self, message: Any):
        self.messages.append(message)


from typing import Type, TypeVar
from .error_info import check_bad_value
from .session import InputSession, Session
from .input_callback import InputCallback
from .callback_data import CallbackDataMapping
from .screen import ReadyScreen, SentScreen

SessionType = TypeVar("SessionType")

class UserData:
    def __init__(self, user_id: int):
        self.user_id = user_id
        self.callback_mapping = CallbackDataMapping()
        self.media_group_id: str | None = None
        self.input_callback: InputCallback | None = None
        self.directory_stack: list[str] = []
        self.screen: SentScreen | None = None
        self.screen_buffer: ReadyScreen | None = None
        self.__sessions: dict[str, Session] = {}
    
    def reset_input_callback(self):
        self.input_callback = None
    
    @property
    def sessions(self):
        return tuple([ item[1] for item in self.__sessions.items() ])
    
    @property
    def input_sessions(self):
        result: list[InputSession] = []
        for session in self.sessions:
            if isinstance(session, InputSession):
                result.append(session)
        return tuple(result)
    
    def add_session(self, session: Session) -> bool:
        check_bad_value(session, Session, self, "session")
        if self.get_session(session.id):
            return False
        
        session.directory_level = len(self.directory_stack)
        try: session.last_directory = self.directory_stack[-1]
        except: pass
        
        self.__sessions[session.id] = session
        return True
        
    def get_session(self, id: str, expected_class: Type[SessionType] = Session
            ) -> SessionType | None:
        return self.__sessions.get(id) # type: ignore
    
    def update_sessions(self):
        new_dir_level = len(self.directory_stack)
        try: last_directory = self.directory_stack[-1]
        except: last_directory = None
        delete = []
        for session in self.sessions:
            if not session.delete_if_level_decreased:
                continue
            if not session.directory_level > new_dir_level:
                continue
            delete.append(session)
            
        for session in self.sessions:
            if not session.delete_if_last_dir_changed:
                continue
            if session.last_directory == last_directory:
                continue
            delete.append(session)
        
        for session in delete:
            self.delete_session(session)
    
    def delete_session(self, session: Session):
        check_bad_value(session, Session, self, "session")
        del self.__sessions[session.id]
    
    def __repr__(self):
        return f"{type(self).__name__}({self.user_id!r})"

class UserDataManager:
    def __init__(self):
        self.__users_data: dict[int, UserData] = {}
        self.users_data = self.__users_data
    
    def get(self, user_id: int) -> UserData:
        user_data = self.__users_data.get(user_id)
        if user_data is None:
            user_data = UserData(user_id)
            self.set(user_id, user_data)
        return user_data
    
    def reset(self, user_id: int) -> None:
        self.set(user_id, UserData(user_id))
    
    def set(self, user_id: int, user_data: UserData):
        self.__users_data[user_id] = user_data



import pathlib
from typing import Self
from telegram import Bot, InputFile, InputMediaAudio
import telegram
from telegram import Message as PTBMessage

from ..button_rows import ButtonRows

from ...callback_data import CallbackDataMapping
from ...message import AudioMessage     as BaseAudioMessage
from ...message import SentAudioMessage     as BaseSentAudioMessage
from .message import HasButtonRows, Message, SentMessage


# TODO: Доработать классы, сейчас они вообще не готовы

class AudioMessage(BaseAudioMessage, HasButtonRows, Message):
    def __init__(self, 
            audio: InputFile | bytes | pathlib.Path | telegram.Audio, 
            caption: str, 
            button_rows: ButtonRows = None, *, 
            parse_mode: str | None = None,
            filename: str = None
        ):
        super().__init__(caption, button_rows, parse_mode)
        self.audio = audio
        self.filename = filename
    
    async def send(self, user_id: int, bot: Bot, mapping: CallbackDataMapping):
        ptb_message = await bot.send_audio(user_id, self.audio, 
            caption = self.caption, 
            reply_markup=self.get_reply_markup(mapping), 
            parse_mode = self.parse_mode)
        return SentAudioMessage(self.audio, ptb_message,
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = ptb_message, 
            filename = self.filename)
    
    def __eq__(self, other: Self):
        return self.caption == other.caption and \
            self.audio == other.audio and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode
    
    def clone(self):
        button_rows = None
        if self.button_rows:
            button_rows = self.button_rows.clone()
        return AudioMessage(self.audio, self.caption, button_rows, 
            parse_mode=self.parse_mode, filename = self.filename)
        
    def transform(self, old: "SentMessage"):
        return SentAudioMessage(self.audio, old.ptb_message, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode, 
            filename = self.filename)

class SentAudioMessage(BaseSentAudioMessage, HasButtonRows, SentMessage):
    def __init__(self, 
            audio: bytes | InputFile | pathlib.Path | telegram.Audio,
            ptb_message: PTBMessage | None,
            caption: str | None = None, 
            button_rows: ButtonRows | None = None, *,
            parse_mode: str | None = None,
            filename: str | None = None
        ):
        super().__init__(caption, button_rows, parse_mode)
        self.audio = audio
        self.ptb_message = ptb_message 
        self.__ptb_message_audio = audio
        self.filename = filename
    
    async def edit(self, bot: Bot, mapping: CallbackDataMapping):
        orig = self.ptb_message
        reply_markup = self.get_reply_markup(mapping)
        if  orig.caption_html == self.caption \
                and orig.reply_markup == reply_markup \
                and self.__ptb_message_audio == self.audio \
                and (
                    not isinstance(self.__ptb_message_audio, telegram.Audio)
                    or self.__ptb_message_audio.file_name == self.filename
                ):
            return
        self.ptb_message = await bot.edit_message_media(
            media=InputMediaAudio(self.audio, self.caption, 
                parse_mode=self.parse_mode, filename=self.filename),
            chat_id=self.ptb_message.chat_id,
            message_id=self.ptb_message.message_id,
            reply_markup = reply_markup)
        self.__ptb_message_audio = self.audio
    
    def __eq__(self, other: Self):
        return self.caption == other.caption and \
            self.audio == other.audio and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode and \
            self.filename == other.filename
    
    def clone(self):
        return self.__class__(self.audio, self.ptb_message, 
            caption = self.caption, 
            button_rows = self.button_rows,
            parse_mode = self.parse_mode, 
            filename = self.filename)
    
    def get_unsent(self):
        return AudioMessage(
            self.audio, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode, 
            filename=self.filename)


import pathlib
from typing import Self
from telegram import Bot, InputFile, InputMediaDocument, InputMediaVideo, Message as PTBMessage
import telegram
from ..button_rows import ButtonRows
from ...callback_data import CallbackDataMapping
from telegram import Message as PTBMessage

from .message import HasButtonRows, Message, SentMessage
from ...message import DocumentMessage  as BaseDocumentMessage
from ...message import SentDocumentMessage  as BaseSentDocumentMessage

class DocumentMessage(BaseDocumentMessage, HasButtonRows, Message):
    def __init__(self, 
            document: bytes | InputFile | pathlib.Path | telegram.Document, 
            caption: str = None, 
            button_rows: ButtonRows = None, *,
            parse_mode: str | None = None, 
            filename: str = None):
        super().__init__(caption, button_rows, parse_mode)
        self.document = document
        self.filename = filename
        
    async def send(self, user_id: int, bot: Bot, mapping: CallbackDataMapping):
        ptb_message = await bot.send_document(user_id, self.document, 
            caption = self.caption, 
            reply_markup=self.get_reply_markup(mapping),
            parse_mode=self.parse_mode,
            filename = self.filename)
        return SentDocumentMessage(
            self.document,
            ptb_message, 
            caption = self.caption,
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode, 
            filename = self.filename)
    
    def __eq__(self, other: Self):
        return self.caption == other.caption and \
            self.document == other.document and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode and \
            self.filename == other.filename
    
    def __repr__(self):
        return f"{type(self).__name__}{{{self.filename}}}({self.caption=!r}, {self.button_rows=!r}, {self.parse_mode=!r})"
    
    def clone(self) -> Self: 
        return self.__class__(self.document, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode, 
            filename = self.filename)
    
    def transform(self, old: "SentMessage"):
        return SentDocumentMessage(
            self.document, old.ptb_message, 
            caption = self.caption,
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode, 
            filename = self.filename)

class SentDocumentMessage(BaseSentDocumentMessage, HasButtonRows, SentMessage):
    def __init__(self, 
            document: bytes | InputFile | pathlib.Path | telegram.Document, 
            ptb_message: PTBMessage, 
            caption: str = None, 
            button_rows: ButtonRows = None, *,
            parse_mode: str | None = None, 
            filename: str = None
        ):
        super().__init__(caption, button_rows, parse_mode)
        self.document = document
        self.ptb_message = ptb_message
        self.__ptb_message_document = document
        self.filename = filename
    
    async def edit(self, bot: Bot, mapping: CallbackDataMapping):
        orig = self.ptb_message
        reply_markup = self.get_reply_markup(mapping)
        if orig.caption_html == self.caption and orig.reply_markup == reply_markup \
                and self.__ptb_message_document == self.document \
                and self.__ptb_message_document.filename == self.filename:
            return
        self.ptb_message = await bot.edit_message_media(
            media=InputMediaDocument(self.document, self.caption, 
                parse_mode=self.parse_mode, filename=self.filename),
            chat_id=self.ptb_message.chat_id,
            message_id=self.ptb_message.message_id,
            reply_markup = reply_markup)
        self.__ptb_message_document = self.document
    
    def __eq__(self, other: Self):
        return self.caption == other.caption and \
            self.document == other.document and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode and \
            self.filename == other.filename
    
    def __repr__(self):
        return f"{type(self).__name__}{{{self.filename}}}({self.caption=!r}, \
{self.button_rows=!r}, {self.parse_mode=!r})"
    
    def clone(self):
        return self.__class__(self.document, self.ptb_message, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode, 
            filename = self.filename)

    def get_unsent(self):
        return DocumentMessage(
            self.document, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode, 
            filename=self.filename)


import pathlib
from typing import Self
from telegram import Bot, InputFile, InputMediaVideo, Message as PTBMessage
import telegram
from ..button_rows import ButtonRows

from ...callback_data import CallbackDataMapping
from telegram import Message as PTBMessage

from .message import HasButtonRows, Message, SentMessage
from ...message import VideoMessage         as BaseVideoMessage
from ...message import SentVideoMessage     as BaseSentVideoMessage

class VideoMessage(BaseVideoMessage, HasButtonRows, Message):
    def __init__(self, 
            video: bytes | InputFile | pathlib.Path | telegram.Video, 
            caption: str = None, 
            button_rows: ButtonRows = None, *,
            parse_mode: str | None = None):
        super().__init__(caption, button_rows, parse_mode)
        self.video = video
        
    async def send(self, user_id: int, bot: Bot, mapping: CallbackDataMapping):
        ptb_message = await bot.send_video(user_id, self.video, 
            caption = self.caption, 
            reply_markup=self.get_reply_markup(mapping),
            parse_mode=self.parse_mode)
        return SentVideoMessage(self.video, ptb_message, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode)
    
    def __eq__(self, other: Self):
        return self.caption == other.caption and \
            self.video == other.video and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode
    
    def __repr__(self):
        return f"{type(self).__name__}({self.caption=!r}, {self.button_rows=!r}, {self.parse_mode=!r})"
    
    def clone(self) -> Self: 
        return self.__class__(self.video, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode)
    
    def transform(self, old: "SentMessage"):
        return SentVideoMessage(self.video, old.ptb_message, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode)

class SentVideoMessage(BaseSentVideoMessage, HasButtonRows, SentMessage):
    def __init__(self, 
            video: bytes | InputFile | pathlib.Path | telegram.Video,
            ptb_message: PTBMessage,
            caption: str = None, 
            button_rows: ButtonRows = None, *,
            parse_mode: str | None = None,
        ):
        super().__init__(caption, button_rows, parse_mode)
        self.video = video
        self.ptb_message = ptb_message
        self.__ptb_message_video = video
    
    async def edit(self, bot: Bot, mapping: CallbackDataMapping):
        orig = self.ptb_message
        reply_markup = self.get_reply_markup(mapping)
        if      orig.caption_html == self.caption \
                and orig.reply_markup == reply_markup \
                and self.__ptb_message_video == self.video:
            return
        self.ptb_message = await bot.edit_message_media(
            media=InputMediaVideo(self.video, self.caption, 
                parse_mode=self.parse_mode),
            chat_id=self.ptb_message.chat_id,
            message_id=self.ptb_message.message_id,
            reply_markup = reply_markup)
        self.__ptb_message_video = self.video
    
    def __eq__(self, other: Self):
        return self.caption == other.caption and \
            self.video == other.video and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode
    
    def __repr__(self):
        return f"{type(self).__name__}({self.caption=!r}, {self.button_rows=!r}, {self.parse_mode=!r})"
    
    def clone(self):
        return self.__class__(self.video, self.ptb_message, 
            caption = self.caption, 
            button_rows = self.button_rows,
            parse_mode = self.parse_mode)

    def get_unsent(self):
        return VideoMessage(
            self.video, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode)


from abc import ABC, abstractmethod
from typing import Any, Callable, TYPE_CHECKING
from .error_info import check_bad_value
if TYPE_CHECKING:
    from .user_data import UserData

class InputCallback(ABC): 
    @abstractmethod
    async def use(self, *, user_id: int, 
        user_data: "UserData", 
        screen_set_by_name: Callable, **kw) -> None: ...


class FuncCallback(InputCallback):
    def __init__(self, function: Callable
        , one_time: bool = True, **kwargs):
        self.function = function
        self.one_time = one_time
        self.kwargs = kwargs
    
    def __call__(self, **kwds):
        return self.function(**self.kwargs, **kwds)
    
    async def use(self, *, user_id: int, 
            user_data: "UserData", 
            screen_set_by_name: Callable, **kw) -> None:
        if self.one_time:
            user_data.input_callback = None
            
        await self.function(user_id=user_id
                , **self.kwargs, **kw)
        

class ScreenCallback(InputCallback):
    def __init__(self, screen_name: str, stack: bool = False):
        self.screen_name = screen_name
        self.stack = stack
        
    async def use(self, *, user_id: int, 
            user_data: "UserData", 
            screen_set_by_name: Callable, **kw) -> None:
        user_data.input_callback = None
        await screen_set_by_name(user_id, self.screen_name,
            self.stack, **kw)
        

def check_pre_post_func(pre: FuncCallback | None, 
                        post: FuncCallback | None, 
                        obj: Any):
    if pre:
        check_bad_value(pre, FuncCallback, obj, "pre_func")
    
    if post:
        check_bad_value(post, FuncCallback, obj, "post_func")


from asyncio import gather
from telegram import Bot
from ..screen import ReadyScreen, SentScreen as BaseSentScreen

class SentScreen(BaseSentScreen):   
    def clone(self) -> "SentScreen":
        return SentScreen(*[message.clone() for message in self.messages])
     
    async def delete(self, bot: Bot):
        tasks = [message.delete(bot)
            for message in self.messages]
        await gather(*tasks)
    
    def get_unsent(self):
        return ReadyScreen(*[
            message.get_unsent() 
            for message in self.messages])


from abc import ABC, abstractmethod
from email import message
from typing import Type, TypeVar, cast
from uuid import uuid4

import telegram

from .error_info import check_bad_value

from .callback_data import CallbackDataMapping
from .screen import ProtoScreen, SentScreen
from .message import Message, SentMessage
from .user_data import UserDataManager
from .screen import ReadyScreen

MsgType = TypeVar('MsgType', bound=Message)
SentMsgType = TypeVar('SentMsgType', bound=SentMessage)

class UserScreen(ABC):
    def __init__(self, user_data: UserDataManager):
        self.user_data = user_data
        self.screen_dict: dict[str, ProtoScreen] = {}
    
    def append_screen(self, screen: ProtoScreen):
        check_bad_value(screen, ProtoScreen, self, "screen")
        if self.screen_dict.get(screen.name) is not None:
            raise KeyError(f"Попытка повторно создать экран с названием {screen.name!r}")
        self.screen_dict[screen.name] = screen
    
    def extend_screen(self, screens: list[ProtoScreen]):
        for screen in screens:
            self.append_screen(screen)
    
    @abstractmethod
    async def clear(self, user_id: int, delete_messages: bool = True): ...
    
    async def set_by_name(self, user_id: int, screen_name: str, 
            stack: bool = True, **kwargs):
        user_data = self.user_data.get(user_id)
        directory_stack = user_data.directory_stack
        
        if not stack:
            if len(directory_stack)==0:
                print(f"{user_id} попытался перейти на экран {screen_name!r} "
                      f"в режиме stack=False, но len(directory_stack) было 0")
                return
            
            if directory_stack[-1] == screen_name: 
                print(f"{user_id} попытался перейти на экран {screen_name!r} "
                      f"но он уже находился на этом экране")
                return 
            user_data.directory_stack[-1] = screen_name
        else:
            if len(directory_stack)==0 or directory_stack[-1] != screen_name:
                directory_stack.append(screen_name)
        
        screen = self.screen_dict.get(screen_name)
        if screen is None:
            raise KeyError(f"Попытка получить экран с названием {screen_name!r}, "
                            "но его не существует")
        evaluated_screen = await screen.evaluate(user_id, sys_user_data=user_data, **kwargs)
        
        await self.set(user_id, evaluated_screen)
    
    async def update(self, user_id: int):
        directory_stack = self.user_data.get(user_id).directory_stack
        if len(directory_stack) != 0:
            await self.set_by_name(user_id, directory_stack[-1])
    
    async def step_back(self, user_id: int, times: int = 1) -> None:
        directory_stack = self.user_data.get(user_id).directory_stack
        for _ in range(times):
            if len(directory_stack) <= 1:
                return
            directory_stack.pop()
        self.user_data.get(user_id).update_sessions()
        await self.set_by_name(user_id, directory_stack[-1])
    
    def get(self, user_id: int) -> SentScreen | None:
        screen = self.user_data.get(user_id).screen
        if screen is None:
            return None
        return screen.clone()
    
    def _map_callback_data(self, user_id: int, screen: ReadyScreen
            ) -> CallbackDataMapping:
        mapping = CallbackDataMapping()
        callback_data_list = screen.get_callback_data()
        for callback_data in callback_data_list:
            uuid = str(uuid4())
            mapping.add(callback_data, uuid)
        self.user_data.get(user_id).callback_mapping = mapping
        return mapping
    
    @abstractmethod
    async def buffer(self, user_id: int): ...
    
    async def unbuffer(self, user_id: int): 
        screen = self.user_data.get(user_id).screen_buffer
        if not screen:
            print(f"у {user_id} нет screen в unbuffer")
            return
        try:
            await self.set(user_id, screen)
        except telegram.error.BadRequest as e:
            print(f"у {user_id} ошибка в unbuffer: {e!r}")
    
    @abstractmethod
    async def set(self, user_id: int, new_screen: ReadyScreen):
        ...
    
    @staticmethod
    def calc_screen_difference(screen1: SentScreen | None, screen2: ReadyScreen,
            msg_type: type[MsgType], sent_msg_type: type[SentMsgType]
        ) -> tuple[list[SentMsgType], list[tuple[SentMsgType, MsgType]], list[MsgType]]:
        
        messages1 = cast(list[SentMsgType], screen1.messages if screen1 else [])
        messages2 = cast(list[MsgType], screen2.messages)
        type_codes = get_type_codes(messages1 + messages2)
        screen1_codes: list[int] = [type_codes[message.category] 
            for message in messages1]
        screen2_codes: list[int] = [type_codes[message.category] 
            for message in messages2]
        
        indices_delete, indices_edit, indices_send = \
            calc_abstract_difference(screen1_codes, screen2_codes)
        
        messages_delete: list[SentMsgType] = [messages1[index]
            for index in indices_delete]
        messages_edit: list[tuple[SentMsgType, MsgType]] = [
            (messages1[from_i],messages2[to_i])
            for from_i, to_i in indices_edit]
        messages_send: list[MsgType] = [messages2[index]
            for index in indices_send]
        return messages_delete, messages_edit, messages_send

SomeMessage = Message | SentMessage

def get_type_codes(messages: list[MsgType | SentMsgType]):
    type_codes = set()
    for message in messages:
        type_codes.add(message.category)
    type_codes = list(type_codes)
    type_codes = [(code, i) for i, code in enumerate(type_codes)]
    return dict(type_codes)

def calc_abstract_difference(start: list[int], end: list[int]
        ) -> tuple[list[int], list[tuple[int, int]], list[int]]:
    indices_delete = []
    indices_edit = []
    indices_send = []
    startn = 0
    for j, enum in enumerate(end):
        if startn >= len(start):
            indices_send.append(j)
            continue
        for i, snum in enumerate(start[startn:], start=startn):
            startn += 1
            if enum == snum:
                indices_edit.append((i, j)) # (from, to)
                break
            else:
                indices_delete.append(i)
        else:
            indices_send = list(range(j, len(end)))
            break
    indices_delete += list(range(startn,len(start)))
    return indices_delete, indices_edit, indices_send


from typing import Self
from telegram import Bot
from telegram import Message as PTBMessage

from ..button_rows import ButtonRows
from ...message import SimpleMessage    as BaseSimpleMessage
from ...message import SentSimpleMessage    as BaseSentSimpleMessage
from ...callback_data import CallbackDataMapping
from .message import HasButtonRows, Message, SentMessage

class SimpleMessage(BaseSimpleMessage, HasButtonRows, Message):
    async def send(self, user_id: int, bot: Bot, mapping: CallbackDataMapping):
        ptb_message = await bot.send_message(user_id, self.text
            , reply_markup=self.get_reply_markup(mapping)
            , parse_mode=self.parse_mode)
        return SentSimpleMessage(
            self.text, self.button_rows, ptb_message, self.parse_mode)
    
    def __eq__(self, other: Self):
        return self.text == other.text and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode
    
    def __repr__(self):
        return f"{type(self).__name__}({self.text!r}, {self.button_rows!r}, {self.parse_mode!r})"
    
    def clone(self):
        button_rows = None
        if self.button_rows:
            button_rows = self.button_rows.clone()
        return SimpleMessage(self.text, button_rows, self.parse_mode)

    def transform(self, old: "SentMessage"):
        return SentSimpleMessage(self.text, self.button_rows,
            old.ptb_message, self.parse_mode)
    
class SentSimpleMessage(BaseSentSimpleMessage, HasButtonRows, SentMessage):
    def __init__(self, text: str, button_rows: ButtonRows
        , ptb_message: PTBMessage, parse_mode: str | None = None):
        super().__init__(text, button_rows, parse_mode)
        self.ptb_message = ptb_message 
    
    async def edit(self, bot: Bot, mapping: CallbackDataMapping):
        orig = self.ptb_message
        reply_markup = self.get_reply_markup(mapping)
        if orig.text_html == self.text and orig.reply_markup == reply_markup:
            return
        self.ptb_message = await bot.edit_message_text(
            text = self.text,
            reply_markup = reply_markup,
            parse_mode = self.parse_mode,
            chat_id=self.ptb_message.chat_id,
            message_id=self.ptb_message.message_id)
    
    def __eq__(self, other: Self):
        return self.text == other.text and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode
            
    def __repr__(self):
        return f"{type(self).__name__}({self.text!r}, {self.button_rows!r})"
    
    def clone(self):
        return SentSimpleMessage(self.text, self.button_rows, self.ptb_message, 
            self.parse_mode)

    def get_unsent(self):
        return SimpleMessage(
              self.text
            , self.button_rows
            , self.parse_mode)


import pathlib
from typing import Self
from telegram import Bot, InputFile, InputMediaPhoto, Message as PTBMessage
import telegram
from ..button_rows import ButtonRows

from ...callback_data import CallbackDataMapping
from ...message import PhotoMessage     as BasePhotoMessage
from ...message import SentPhotoMessage     as BaseSentPhotoMessage
from .message import HasButtonRows, Message, SentMessage

class PhotoMessage(BasePhotoMessage, HasButtonRows, Message):
    def __init__(self,
            photo: bytes | InputFile | pathlib.Path | telegram.PhotoSize, 
            caption: str = None, 
            button_rows: ButtonRows = None, *,
            parse_mode: str | None = None):
        super().__init__(caption, button_rows, parse_mode)
        self.photo = photo
        
    async def send(self, user_id: int, bot: Bot, mapping: CallbackDataMapping):
        ptb_message = await bot.send_photo(user_id, self.photo
            , self.caption, reply_markup=self.get_reply_markup(mapping)
            , parse_mode=self.parse_mode)
        return SentPhotoMessage(self.photo, 
            ptb_message = ptb_message, 
            caption = self.caption,
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode)
    
    def __eq__(self, other: Self):
        return self.caption == other.caption and \
            self.photo == other.photo and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode
    
    def __repr__(self):
        return f"{type(self).__name__}({self.caption=!r}, {self.button_rows=!r}, {self.parse_mode=!r})"
    
    def clone(self) -> Self: 
        return self.__class__(self.photo, self.caption, self.button_rows, 
            self.parse_mode)
    
    def transform(self, old: "SentMessage"):
        return SentPhotoMessage(
            self.photo, 
            ptb_message = old.ptb_message, 
            caption = self.caption,
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode)


class SentPhotoMessage(BaseSentPhotoMessage, HasButtonRows, SentMessage):
    def __init__(self, 
            photo: bytes | InputFile | pathlib.Path | telegram.PhotoSize, 
            ptb_message: PTBMessage, 
            caption: str = None, 
            button_rows: ButtonRows = None, *, 
            parse_mode: str | None = None,
        ):
        super().__init__(caption, button_rows, parse_mode)
        self.photo = photo
        self.ptb_message = ptb_message
        self.__ptb_message_photo = photo
    
    async def edit(self, bot: Bot, mapping: CallbackDataMapping):
        orig = self.ptb_message
        reply_markup = self.get_reply_markup(mapping)
        if orig.caption_html == self.caption and orig.reply_markup == reply_markup \
                and self.__ptb_message_photo == self.photo:
            return
        self.ptb_message = await bot.edit_message_media(
            media=InputMediaPhoto(self.photo, self.caption, self.parse_mode),
            chat_id=self.ptb_message.chat_id,
            message_id=self.ptb_message.message_id,
            reply_markup = reply_markup)
        self.__ptb_message_photo = self.photo
    
    def __eq__(self, other: Self):
        return self.caption == other.caption and \
            self.photo == other.photo and \
            self.button_rows == other.button_rows and \
            self.parse_mode == other.parse_mode
    
    def __repr__(self):
        return f"{type(self).__name__}({self.caption=!r}, {self.button_rows=!r}, {self.parse_mode=!r})"
    
    def clone(self):
        return self.__class__(self.photo, self.ptb_message, 
            caption = self.caption, 
            button_rows = self.button_rows,
            parse_mode = self.parse_mode)

    def get_unsent(self):
        return PhotoMessage(
            self.photo, 
            caption = self.caption, 
            button_rows = self.button_rows, 
            parse_mode = self.parse_mode)


from abc import ABC, abstractmethod
from telegram import Bot, Message as PTBMessage
import telegram

from ..button_rows import ButtonRows
from ...callback_data import CallbackDataMapping
from ...message import Message as BaseMessage
from ...message import SentMessage as BaseSentMessage
from ...message import HasButtonRows as BaseHasButtonRows

class HasButtonRows(BaseHasButtonRows):
    def __init__(self):
        self.button_rows: ButtonRows | None = None
        
    def get_reply_markup(self, mapping: CallbackDataMapping):
        if self.button_rows:
            return self.button_rows.to_reply_markup(mapping)
        return None


class Message(BaseMessage, ABC):
    @abstractmethod
    def transform(self, old: "SentMessage") -> "SentMessage":
        ...
    

class SentMessage(BaseSentMessage):
    def __init__(self):
        self.ptb_message: PTBMessage = None
    
    async def delete(self, bot: Bot):
        try:
            await bot.delete_message(
                chat_id=self.ptb_message.chat_id,
                message_id=self.ptb_message.message_id)
        except telegram.error.BadRequest as e:
            print(f"{self} не получилось удалить: {e!r}")
    
    @abstractmethod
    async def edit(self, bot: Bot, mapping: CallbackDataMapping): ...


from typing import Any
from ..session import InputSession as BaseInputSession
from telegram import Message as TgMessage


class InputSession(BaseInputSession):
    def __init__(self, id: str, delete_if_level_decreased=True, delete_if_last_dir_changed=False, 
            add_new_messages=True, may_pop_last_input=True):
        super().__init__(id, delete_if_level_decreased, delete_if_last_dir_changed, 
            add_new_messages, may_pop_last_input)
        
        self.messages: list[TgMessage] = []
    
    def append(self, message: TgMessage):
        self.messages.append(message)
        
    def __repr__(self):
        texts = []
        for message in self.messages:
            texts.append(message.text)
        return f"{type(self).__name__}(id={self.id!r}, texts={texts!r})"


from abc import abstractmethod, ABC
from typing import Callable, Self
from .screen import DynamicScreen
from .input_callback import FuncCallback, InputCallback, ScreenCallback
from .callback_data import CallbackData
from .user_data import UserDataManager
from .user_screen import UserScreen

class BotManager(ABC):
    def __init__(self):
        self.system_user_data: UserDataManager
        self.screen: UserScreen
        
    def config_delete_old_messages(self, user_id: int, **kwargs):
        input_callback = self.get_system_user_data(user_id).input_callback
        if input_callback is not None:
            return False
        return True
    
    @abstractmethod
    def build(self) -> Self: ...
        # user_datam = UserDataManager()
        # screen = UserScreen(user_datam)
        # self.system_user_data = user_datam
        # self.screen = screen
        # return self
    
    @abstractmethod
    def add_handlers(self): ...
    
    def get_system_user_data(self, user_id: int):
        return self.system_user_data.get(user_id)

    @abstractmethod
    def get_message_handler(self): ...

    async def _handle_message(self, user_id: int, **kwargs):
        user_data = self.get_system_user_data(user_id)
        delete_old: bool = self.config_delete_old_messages(user_id, **kwargs)
        if delete_old:
            await self.delete_message(**kwargs)
        
        input_callback = user_data.input_callback
        if input_callback is None:
            return
        
        await self.screen.clear(user_id, delete_old)
        
        for session in user_data.input_sessions:
            if not session.add_new_messages:
                continue
            message = kwargs["message"]
            session.append(message)
        
        await input_callback.use(
            user_id=user_id, user_data=user_data,
            screen_set_by_name=self.screen.set_by_name,
            **kwargs
        )
            

    @abstractmethod
    def get_callback_query_handler(self): ...
    
    @abstractmethod
    async def delete_message(self, message): ...
    
    async def mapping_key_error(self, user_id: int): ...
    
    async def _handle_callback_query(self, user_id: int, query_data: str, **kwargs):
        sud = self.get_system_user_data(user_id)
        mapping = sud.callback_mapping
        data: CallbackData | None = mapping.get_by_uuid(query_data)
        if data is None:
            await self.mapping_key_error(user_id)
            return
        
        await data.use(user_id=user_id,
            input_sessions=sud.input_sessions,
            screen_set_by_name=self.screen.set_by_name,
            screen_step_back=self.screen.step_back,
            reset_input_callback=sud.reset_input_callback,
            update_sessions=sud.update_sessions,
            **kwargs)
    
    def dynamic_screen(self, name: str | None = None):
        def decorator(func: Callable):
            nonlocal name
            if name is None:
                name = func.__name__
            self.screen.append_screen(DynamicScreen(name, func))
        return decorator


    
    



from telegram import Update, Bot
from telegram import Message as TgMessage
from telegram.ext import Application, CallbackQueryHandler, MessageHandler

from ..bot_manager import BotManager as BaseBotManager
from .user_screen import UserScreen
from ..user_data import UserDataManager

class BotManager(BaseBotManager):
    def __init__(self, application: Application):
        super().__init__()
        self.bot: Bot = application.bot
        self.application = application
    
    def build(self):
        user_data = UserDataManager()
        screen = UserScreen(user_data, self.bot)
        self.system_user_data = user_data
        self.screen = screen
        return self
    
    def get_callback_query_handler(self):
        async def callback(update: Update, context):
            user_id = update.callback_query.from_user.id
            query_data = update.callback_query.data
            await self._handle_callback_query(user_id, query_data, 
                update=update, context=context)
            await update.callback_query.answer()
        return CallbackQueryHandler(callback)

    def get_message_handler(self):
        async def callback(update: Update, context):
            user_id = update.message.from_user.id
            await self._handle_message(user_id, update=update
                , message=update.message, context=context)
            
        return MessageHandler(None, callback)
    
    async def delete_message(self, message: TgMessage, **kwargs):
        await message.delete()
    
    def add_handlers(self):
        """Эта функция должна вызываться ПОСЛЕ 
        регистрации /start или подобных команд"""
        self.application.add_handlers([
            self.get_callback_query_handler(),
            self.get_message_handler()
        ])


from uuid import uuid4
from telegram import InlineKeyboardButton, InlineKeyboardMarkup, WebAppInfo

from tg_bot_screen.callback_data import CallbackData

from ..callback_data import CallbackDataMapping
from ..button_rows import ButtonRows as BaseButtonRows
from ..button_rows import Button as BaseButton
from ..button_rows import ButtonRow as BaseButtonRow

class ButtonRows(BaseButtonRows):
    def to_reply_markup(self, mapping: CallbackDataMapping
            ) -> InlineKeyboardMarkup:
        result = []
        for row in self.rows:
            row_list = []
            for button in row.buttons:
                button: Button
                row_list.append(button.to_inline_button(mapping))
            result.append(row_list)
        return InlineKeyboardMarkup(result)
    
    def clone(self) -> "ButtonRows":
        return ButtonRows(*[row.clone() for row in self.rows])

class Button(BaseButton):
    def __init__(self, text: str, callback_data: CallbackData, 
                 url: str | None = None, web_app: WebAppInfo|None=None):
        super().__init__(text, callback_data, url, web_app)
        self.web_app = web_app
        
        
    def to_inline_button(self, mapping: CallbackDataMapping
            ) -> InlineKeyboardButton:
        if self.web_app:
            uuid = None
        else:
            uuid = mapping.get_by_callback(self.callback_data)
        
        return InlineKeyboardButton(self.text
            , callback_data = uuid, url = self.url, web_app=self.web_app)
    
    def clone(self):
        return Button(self.text, self.callback_data.clone(), 
                      self.url, self.web_app)
        

class ButtonRow(BaseButtonRow): ...


from typing import Any, Callable, Type

def get_value_error(arg_value: Any, expected_type: Type | Callable, 
                    obj: Any, arg_name: str):
    return ValueError(f"У {type(obj).__name__} аргумент "
            f"{arg_name}={arg_value!r} неправильного типа "
            f"{type(arg_value).__name__} "
            f"(Ожидался {expected_type.__name__})")

def check_callable(arg_value: Any, obj: Any, arg_name: str):
    if not callable(arg_value):
        raise get_value_error(arg_value, Callable, obj, arg_name)

def check_bad_value(arg_value: Any, expected_type: Type, 
                    obj: Any, arg_name: str): 
    if not isinstance(arg_value, expected_type):
        raise get_value_error(arg_value, expected_type, obj, arg_name)

def check_bad_text(arg_value: str, obj: Any, arg_name: str):
    check_bad_value(arg_value, str, obj, arg_name)
        
def check_bad_text_and_len(arg_value: str, obj: Any, arg_name: str):
    check_bad_text(arg_value, obj, arg_name)
    if len(arg_value) == 0:
        raise ValueError(f"У {obj!r} аргумент {arg_name}={arg_value!r} "
                         f"не может быть {""!r}")




from ..callback_data import CallbackData, RunFunc, GoToScreen, StepBack
from ..input_callback import InputCallback, FuncCallback, ScreenCallback
from ..screen import ReadyScreen, StaticScreen, DynamicScreen, SentScreen, \
    ProtoScreen, StaticScreen, DynamicScreen
from ..message import Message, SentMessage
from ..session import Session

from .bot_manager import BotManager
from .button_rows import ButtonRows, ButtonRow, Button
from .messages.message import Message, SentMessage
from .messages.simple_message import SimpleMessage, SentSimpleMessage
from .messages.photo_message import PhotoMessage, SentPhotoMessage
from .messages.video_message import VideoMessage, SentVideoMessage
from .messages.audio_message import AudioMessage, SentAudioMessage
from .messages.document_message import DocumentMessage, SentDocumentMessage
from .messages.video_note_message import VideoNoteMessage, SentVideoNoteMessage

from .screen import SentScreen
from .user_screen import UserScreen
from .user_data import UserData, UserDataManager
from .session import InputSession


from abc import ABC, abstractmethod
from typing import Any, Self

from .callback_data import CallbackData
from .button_rows import ButtonRows

class CanBeEdited(ABC):
    @abstractmethod
    async def edit(self) -> "CanBeEdited": ...

class HasButtonRows(ABC):
    def __init__(self):
        self.button_rows: ButtonRows | None = None
        
    @abstractmethod
    def get_reply_markup(self, *args, **kwargs):
        ...
    
    def get_callback_data(self) -> list[CallbackData]:
        if self.button_rows is None:
            return []
        return self.button_rows.get_callback_data()

class SentMessage(ABC):
    def __init__(self):
        self.text: str
        self.category: str
    
    @abstractmethod
    async def delete(self): ...
    
    @abstractmethod
    def __eq__(self, other: object) -> bool: 
        ...
    
    @abstractmethod
    def clone(self) -> Self: ...
    
    @abstractmethod
    def get_unsent(self): ...

class Message(ABC):
    def __init__(self):
        self.category: str | None = None
    
    @abstractmethod
    async def send(self, user_id: int) -> SentMessage: ...
    
    @abstractmethod
    def __eq__(self, other: object) -> bool: 
        ...
    
    @abstractmethod
    def clone(self) -> Self: ...

class MediaMessage(ABC):
    def __init__(self, caption: str, button_rows: ButtonRows | None = None, 
                parse_mode: str | None = None):
        self.parse_mode = parse_mode
        self.caption = caption
        self.button_rows = button_rows
        self.category = "media"
    
    def __new__(cls, *args, **kwargs):
        assert \
            cls is not MediaMessage, \
            "Нельзя создавать экземпляры MediaMessage напрямую"
        return super().__new__(cls)


class AudioMessage(MediaMessage, Message): ...

class DocumentMessage(MediaMessage, Message): ...

class SimpleMessage(Message):
    def __init__(self, text: str, button_rows: ButtonRows | None = None, 
                    parse_mode: str | None = None):
        self.parse_mode = parse_mode
        self.text = text
        self.button_rows = button_rows
        self.category = "simple"

class PhotoMessage(MediaMessage, Message): ...

class VideoMessage(MediaMessage, Message): ...

class VideoNoteMessage(Message):
    def __init__(self, caption: str, button_rows: ButtonRows | None = None):
        self.caption = caption
        self.button_rows = button_rows
        self.category = "video_note"

class SentAudioMessage(MediaMessage, SentMessage, CanBeEdited): ...

class SentDocumentMessage(MediaMessage, SentMessage, CanBeEdited): ...

class SentSimpleMessage(SentMessage, CanBeEdited):
    def __init__(self, text: str, button_rows: ButtonRows | None = None, 
            parse_mode: str | None = None):
        self.parse_mode = parse_mode
        self.text = text
        self.button_rows = button_rows
        self.category = "simple"

class SentPhotoMessage(MediaMessage, SentMessage, CanBeEdited): ...

class SentVideoMessage(MediaMessage, SentMessage, CanBeEdited): ...

class SentVideoNoteMessage(SentMessage):
    def __init__(self, caption: str, button_rows: ButtonRows | None = None):
        self.caption = caption
        self.button_rows = button_rows
        self.category = "video_note"





from ..user_data import UserData as BaseUserData, \
    UserDataManager as BaseUserDataManager
from .screen import SentScreen
from .session import InputSession

class UserData(BaseUserData):
    def __init__(self, user_id: int):
        super().__init__(self, user_id)
        self.screen: SentScreen = None
    
    @property
    def input_sessions(self):
        result: list[InputSession] = []
        for session in self.__sessions:
            if isinstance(session, InputSession):
                result.append(session)
        return tuple(result)
        
class UserDataManager(BaseUserDataManager):
    def __init__(self):
        self.__users_data: dict[int, UserData] = {}
        self.users_data = self.__users_data
    
    def get(self, user_id: int) -> UserData:
        user_data = self.__users_data.get(user_id)
        if user_data is None:
            user_data = UserData(user_id)
            self.set(user_id, user_data)
        return user_data
    
    def reset(self, user_id: int) -> None:
        self.set(user_id, UserData(user_id))
    
    def set(self, user_id: int, user_data: UserData):
        self.__users_data[user_id] = user_data
